/*  WB-tree File Based Associative String Data Base System. */
/*  Copyright (c) 1991, 1992, 1993, 2000 Free Software Foundation, Inc. */
/*  */
/*  This program is free software; you can redistribute it and/or modify */
/*  it under the terms of the GNU General Public License as published by */
/*  the Free Software Foundation; either version 2, or (at your option) */
/*  any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/*  GNU General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU General Public License */
/*  along with this software; see the file COPYING.  If not, write to */
/*  the Free Software Foundation, 59 Temple Place, Suite 330, Boston, */
/*  MA 02111, USA. */
/*  */
/*  As a special exception, the Free Software Foundation gives permission */
/*  for additional uses of the text contained in its release of GUILE. */
/*  */
/*  The exception is that, if you link the GUILE library with other files */
/*  to produce an executable, this does not by itself cause the */
/*  resulting executable to be covered by the GNU General Public License. */
/*  Your use of that executable is in no way restricted on account of */
/*  linking the GUILE library code into it. */
/*  */
/*  This exception does not however invalidate any other reasons why */
/*  the executable file might be covered by the GNU General Public License. */
/*  */
/*  This exception applies only to the code released by the */
/*  Free Software Foundation under the name GUILE.  If you copy */
/*  code from other Free Software Foundation releases into a copy of */
/*  GUILE, as the General Public License permits, the exception does */
/*  not apply to the code that you add in this way.  To avoid misleading */
/*  anyone as to the status of such modified files, you must delete */
/*  this exception notice from them. */
/*  */
/*  If you write modifications of your own for GUILE, it is your choice */
/*  whether to permit this exception to apply to your modifications. */
/*  If you do not wish that, delete this exception notice. */


/* A @dfn{segment} is a self-contained collection of B-trees.  Each is */
/* usually associated with its own (disk) file.  WB functions refer to */
/* segments by integers between 0 and @code{num_segs} - 1. */
/* @code{num_segs} is defined in @url{defs.scm} and derived @url{defs.h}.  The */
/* default value of @code{num_segs} is 10; The minimum value of @code{num_segs} is 1. */

#define num_segs 0xa

/* @cindex flc */
/* Each segment (file-based or not) maintains a @dfn{freelist} of disk */
/* blocks.  In a newly created segment, nearly all of its block numbers */
/* are in the freelist. */
/* @code{flc_len} must be larger than 2 times the maximum number of blocks */
/* which would ever be needed for a freelist split.  The minimum for @code{flc_len} */
/* is 10. */

#define flc_len 0x14


/* amount to increase the ENT-TAB by when allocating buffers. */

#define ent_tab_inc 0x200

/*  ERROR Return Codes */


#define success 0

#define notpres -1

#define terminated -2

#define retryerr -10

#define argerr -15

#define noroom -20

#define typerr -30

#define ioerr -40

#define strangerr -45

#define unkerr -90

#define maxerr -100

/* Return @var{x} if a valid error code (-1 @dots{} @var{MAXERR}); */
/* else 0. */

#define err_P(x) (((1)) && (0 > (x)) && ((x)>=(maxerr)) && (x))


#define realerr_P(x) (((1)) && ((x)<=(retryerr)) && ((x)>=(maxerr)) && (x))

/* Not @code{@code{err_P}}. */

#define success_P(x) (!(err_P(x)))

/*  BLK parameters */

/*  The IDs are 4 byte numbers identifying this block, the root of */
/*  this tree, and the next in the chain. */

#define blk_id_pos 0

#define blk_top_id_pos 4

#define blk_nxt_id_pos 8

#define blk_time_pos 0xc
/*  blk-end-pos is position (stored in 2 bytes) of first free byte */

#define blk_end_pos 0x10

#define blk_level_pos 0x12

#define blk_typ_pos 0x13

#define blk_data_start 0x14


#define blk_id(blk) (str2long(blk, blk_id_pos))

#define blk_top_id(blk) (str2long(blk, blk_top_id_pos))

#define blk_nxt_id(blk) (str2long(blk, blk_nxt_id_pos))

#define blk_time(blk) (str2long(blk, blk_time_pos))

#define blk_end(blk) (str2short(blk, blk_end_pos))

#define blk_level(b) ((unsigned)(b[blk_level_pos]))

#define blk_typ(b) (b[blk_typ_pos])

#define blk_typ_P(b, typ) ((b[blk_typ_pos])==(typ))


#define blk_set_id(blk, id) {long2str(blk, blk_id_pos, id);}

#define blk_set_top_id(blk, id) {long2str(blk, blk_top_id_pos, id);}

#define blk_set_nxt_id(blk, id) {long2str(blk, blk_nxt_id_pos, id);}

#define blk_set_time(blk, tim) {long2str(blk, blk_time_pos, tim);}

#define blk_set_end(blk, pos) {short2str(blk, blk_end_pos, pos);}

#define blk_set_level(b, level) {b[blk_level_pos] = (level);}

#define blk_set_typ(b, typ) {b[blk_typ_pos] = typ;}


#define leaf ((unsigned)('0'))


#define dir_typ 'D'

#define ind_typ 'T'

#define seq_typ 'S'

#define frl_typ 'F'


#define wcb_sap 1

#define wcb_sar 2

#define wcb_sac 4

#define wcb_fac 8


#define wcb_sap_P(wcb) ((wcb_sap)&(wcb))

#define wcb_sar_P(wcb) ((wcb_sar)&(wcb))

#define wcb_sac_P(wcb) ((wcb_sac)&(wcb))

#define wcb_fac_P(wcb) ((wcb_fac)&(wcb))


#define end_of_chain -1

#define start_of_chain -2


#define field_len(blk, pos) ((unsigned)(blk[pos]))


#define set_field_len(blk, pos, len) {blk[pos] = (len);}

/*  This is dangerous.  At the moment all occurences of next-field */
/*  have simple expressions for the second argument. */


#define next_field(blk, pos) ((field_len(blk, pos))+(pos)+1)


#define next_cnvpair(blk, pos) (next_field(blk, next_field(blk, 1+(pos))))


#define leaf_P(blk) ((blk_level(blk))==(leaf))

/*  LCK and ENT tables */

/*  If you change this change amnesia-ent! */
/*  This depends on seg never being less than -1 */

#define hash2int(seg, num) ((((seg)*0x61)+(num)+((num_buks)*(1+(0x61/(num_buks)))))%(num_buks))

/*  Called with SEG-LCK locked. */
/*  If you don't know what you are doing. DON'T DO IT! */
/*  Compute inverse hash function so that ent can still be found. */

#define amnesia_ent(ent) {ent_set_id(ent, hash2int(1+(ent_seg(ent)), ent_id(ent)));ent_set_dty(ent, 0);ent_set_pus(ent, 0);if (ent_blk(ent))if (blk_typ_P(ent_blk(ent), dir_typ))blk_set_typ(ent_blk(ent), ind_typ);ent_set_seg(ent, -1);ent_set_age(ent, 0x80);}


#define same_buk_P(a_seg, a_num, b_seg, b_num) ((hash2int(a_seg, a_num))==(hash2int(b_seg, b_num)))


#define get_buk(seg, blk_num) (buk_tab[hash2int(seg, blk_num)])

/*  doesnt wait, ie, returns #F if busy */

#define get_buk_lck(seg, blk_num) (try_lck(lck_tab[hash2int(seg, blk_num)]))


#define get_buk_wait(seg, blk_num) (lck(lck_tab[hash2int(seg, blk_num)]),buk_tab[hash2int(seg, blk_num)])


#define rel_buk(seg, blk_num) {unlck(lck_tab[hash2int(seg, blk_num)]);}

/*  SET-BUK! assumes BUK is already lcked by caller */

#define set_buk(seg, blk_num, ent) {buk_tab[hash2int(seg, blk_num)] = ent;}


#define accread (1)

#define accwrite (2)

#define accpend (3)

/*  Routines for finding the appropriate BLK for an operation. */
/*  PACKETs used to return multiple values from chain-find. */
/*  and various other operations */


#define pkt_size 6


#define match_type(p) (p[0])

#define match_pos(p) (p[1])

#define key_pos(p) (p[2])

#define prev_match_pos(p) (p[3])

#define blk_to_cache(p) (p[4])

#define success_code(p) (p[5])


#define set_match_type(p, v) {p[0] = v;}

#define set_match_pos(p, v) {p[1] = v;}

#define set_key_pos(p, v) {p[2] = v;}

#define set_prev_match_pos(p, v) {p[3] = v;}

#define set_blk_to_cache(p, v) {p[4] = v;}

#define set_success_code(p, v) {p[5] = v;}


#define pack(p, type, b_pos, k_pos, p_pos) {set_match_type(p, type);set_match_pos(p, b_pos);set_key_pos(p, k_pos);set_prev_match_pos(p, p_pos);}


#define pastp (4)
			/* match(new-key, after-key) > repeatcount(after-key) */

#define qpastp (5)
			/* match(new-key, after-key) <= repeatcount(after-key) */

#define match (6)

#define matchend (7)

#define pastend (8)


#define rem_scan -1

#define count_scan 0

#define modify_scan 1


#define skey_count match_pos

#define set_skey_count set_match_pos

#define skey_len key_pos

#define set_skey_len set_key_pos


#define seg_port(seg) (segd_port(segd_tab[seg]))

#define seg_bsiz(seg) (segd_bsiz(segd_tab[seg]))

#define seg_used(seg) (segd_used(segd_tab[seg]))

#define seg_str(seg) (segd_str(segd_tab[seg]))

#define seg_rt_han(seg) (segd_rt_han(segd_tab[seg]))

#define seg_fl_han(seg) (segd_fl_han(segd_tab[seg]))

#define seg_lck(seg) (segd_lck(segd_tab[seg]))

#define seg_fck(seg) (segd_fck(segd_tab[seg]))

#define seg_flc_len(seg) (segd_flc_len(segd_tab[seg]))

#define seg_flc(seg) (segd_flc(segd_tab[seg]))


#define seg_set_port(seg, port) {segd_set_port(segd_tab[seg], port);}

#define seg_set_bsiz(seg, bsiz) {segd_set_bsiz(segd_tab[seg], bsiz);}

#define seg_set_used(seg, used) {segd_set_used(segd_tab[seg], used);}

#define seg_set_str(seg, str) {segd_set_str(segd_tab[seg], str);}

#define seg_set_flc_len(seg, flc_len) {segd_set_flc_len(segd_tab[seg], flc_len);}

#define seg_set_flc(seg, flc) {segd_set_flc(segd_tab[seg], flc);}
